
#include "cuda_runtime.h"
#include "device_launch_parameters.h"

#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <iostream>
#include <string>

/*************************
 * @brief: Device Pointers
 *************************/
struct Dev_t {
    int* Vec_A;
};

/***********************
 * @brief: Host Pointers
 **********************/
struct Host_t {
    int* Vec_A;
};



/**************************************
 * @brief: Simple Bitonic Kernel Sorter 
 *************************************/
__global__ void Bitonic_Kernel(int*Global,unsigned int N,unsigned int Stage,unsigned int Dist){
    unsigned int tix = threadIdx.x + blockIdx.x * blockDim.x;
    
    if (tix < (N / 2)) {
        unsigned int StagePw_C     = (1 << (Stage + 0)); // 2^ (Stage + 0)
        unsigned int StagePw_N     = (1 << (Stage + 1)); // 2^ (Stage + 1)
        unsigned int ThreadBaseInd = (tix % Dist) + ((tix / Dist) << 1) *  Dist; // mod(tid,Dist) + (intdev(tid,Dist) * 2 * Dist)

        // ------------------
        // -- Data Indexes --
        unsigned int Left_Ind      = ThreadBaseInd; 
        unsigned int Right_Ind     = ThreadBaseInd + Dist;

        // --------------------------------------
        // ------ Load From Global Memory -------
        int Left = Global[Left_Ind];
        int Right = Global[Right_Ind];

        // -------------------------------------------------------------
        // ------- Compute Sequence Direction (Rising / Falling) -------
        unsigned int Sequence_Dir = (tix % StagePw_N);
        if (Sequence_Dir < StagePw_C){
            // ------------------------------
            // ------- Rising Sequence ------
            if (Right < Left) {
                Global[Right_Ind] = Left;
                Global[Left_Ind] = Right;
            }

        }else {
            // --------------------------------
            // ------- Falling Sequence -------
            if (Right > Left) {
                Global[Right_Ind] = Left;
                Global[Left_Ind] = Right;
            }

        }

    }
}

/*******************************************
 * @brief: Verifies that sequence is Bitonic 
 ******************************************/
void CPU_Bitonic_Verification(int *Data,int N) {
    int SGN = 0;
    int Switches = 0;
    int FirstSample = 1;
    for (int i = 1; i < N; i++) {
        int Diff = Data[i] - Data[i - 1];

        // ---------------------------------------------------
        // ------- Ignore Difference of 0 - Same Order -------
        
        // -------------------------------
        // ------- Rising Sequence -------
        if (Diff > 0) {
            if (!FirstSample) {
                if (SGN != 1) {
                    Switches++;
                    SGN = 1;
                }
            }else {
                // ------------------------------------
                // ------- First Rising Sample --------
                FirstSample = 0;
                SGN = 1; 
            }
        
        }
        // --------------------------------
        // ------- Falling Sequence -------
        if (Diff < 0) {
            if (!FirstSample) {
                if (SGN != 0) {
                    Switches++;
                    SGN = 0;
                }
            }else {
                // ------------------------------------
                // ------- First Rising Sample --------
                FirstSample = 0;
                SGN = 0;
            }
        }
    }

    std::cout << "CPU Verification - Switches (Should be 1!): " << std::to_string(Switches) << std::endl;
}

int main(int argc, char** argv){

    int N = 1024*1024*64;

    // ----------------------------------------
    // ------- Compute Amount of Stages -------
    unsigned int Stages = static_cast<unsigned int>(std::ceil(std::log2(static_cast<double>(N)))) - 1;
    Dev_t Dev;
    Host_t Host;
    cudaEvent_t Start;
    cudaEvent_t Stop;

    float Elapsed_Time;
    cudaEventCreate(&Start);
    cudaEventCreate(&Stop);

    // ------------------------------------
    // -------- Allocate GPU Memory -------
    cudaMalloc((void**)&Dev.Vec_A, N * sizeof(int));

    // ----------------------------------
    // ------- Allocate CPU Memory ------
    cudaHostAlloc((void**)&Host.Vec_A, N * sizeof(int), cudaHostAllocDefault);

    // ----------------------------------
    // ------ Randomize Input Data ------
    srand(time(NULL));
    for (int i = 0; i < N; i++) {
        Host.Vec_A[i] = rand() % N;
    }

    // ---------------------------------------
    // ------- Setup Kernel Dimensions -------
    int Threads_Per_Block = 512;
    int Blocks_Per_Grid = (N / 2) / Threads_Per_Block;
    dim3 Threads = dim3(Threads_Per_Block, 1, 1);
    dim3 Blocks = dim3(std::max(1, Blocks_Per_Grid), 1, 1);

    // ------------------------------
    // ------- Copy to Device -------
    cudaMemcpy(Dev.Vec_A, Host.Vec_A, N * sizeof(int), cudaMemcpyHostToDevice);

    // ---------------------------------------
    // ------- Do the Bitonic Sorting --------
    cudaEventRecord(Start);
    for (unsigned int Stage = 0; Stage < Stages; Stage++) {

        // --------------------------------
        // ------- Compute Substages ------
        unsigned int Dist = (1 << Stage);
        for (unsigned int SubStage = 0; SubStage <= Stage; SubStage++) {
            Bitonic_Kernel << <Blocks, Threads >> > (Dev.Vec_A, N, Stage, Dist);
            Dist /= 2;
        }
    }
    cudaEventRecord(Stop);
    cudaEventSynchronize(Stop);
    cudaEventElapsedTime(&Elapsed_Time, Start, Stop);

    // --------------------------------
    // ------- Copy From Device -------
    cudaMemcpy(Host.Vec_A, Dev.Vec_A, N * sizeof(int), cudaMemcpyDeviceToHost);

    // -----------------------------
    // ------- Verify Results ------
    CPU_Bitonic_Verification(Host.Vec_A, N);

    std::cout << "Elapsed Time      : " << std::to_string(Elapsed_Time) << " ms" << std::endl;

    // ---------------------------------
    // ------- Cleanup Resources -------
    cudaEventDestroy(Start);
    cudaEventDestroy(Stop);
    cudaFreeHost(Host.Vec_A);
    cudaFree(Dev.Vec_A);


    return 0;
}